'use strict'

const path = require('path')
const nock = require('../.')
const url = require('url')
const http = require('http')
const https = require('https')
const { test } = require('tap')
const mikealRequest = require('request')
const superagent = require('superagent')
const needle = require('needle')
const restify = require('restify-clients')
const domain = require('domain')
const hyperquest = require('hyperquest')
const async = require('async')
const got = require('got')
const lolex = require('lolex')
const debug = require('debug')
const sinon = require('sinon')

const textFile = path.join(__dirname, '..', 'assets', 'reply_file_1.txt')
const binaryFile = path.join(__dirname, '..', 'assets', 'reply_file_2.txt.gz')

nock.enableNetConnect()

const globalCount = Object.keys(global).length
const acceptableLeaks = [
  '_key',
  '__core-js_shared__',
  'fetch',
  'Response',
  'Headers',
  'Request',
]

test('invalid or missing method parameter throws an exception', t => {
  t.throws(() => nock('https://example.com').intercept('/somepath'), {
    message: 'The "method" parameter is required for an intercept call.',
  })
  t.end()
})

test('double activation throws exception', t => {
  nock.restore()
  t.false(nock.isActive())

  nock.activate()
  t.true(nock.isActive())

  t.throws(() => nock.activate(), { message: 'Nock already active' })

  t.true(nock.isActive())

  t.end()
})

test("when request's content-type is json: reply callback's requestBody should automatically parse to JSON", async t => {
  const requestBodyFixture = {
    id: 1,
    name: 'bob',
  }

  const scope = nock('http://service')
    .post('/endpoint')
    .reply(200, (uri, requestBody) => {
      t.deepEqual(requestBody, requestBodyFixture)
      return 'overwrite'
    })

  const { body } = await got.post('http://service/endpoint', {
    headers: { 'Content-Type': 'application/json' },
    body: JSON.stringify(requestBodyFixture),
  })

  t.equal(body, 'overwrite')
  scope.done()
})

test("when the path doesn't include a leading slash it raises an error", function(t) {
  t.plan(1)
  t.throws(() => nock('http://example.test').get('no-leading-slash'))
})

test("when request has no content-type header: reply callback's requestBody should not automatically parse to JSON", async t => {
  const requestBodyFixture = {
    id: 1,
    name: 'bob',
  }

  const scope = nock('http://service')
    .post('/endpoint')
    .reply(200, (uri, requestBody) => {
      t.deepEqual(requestBody, JSON.stringify(requestBodyFixture))
      return 'overwrite'
    })

  const { body } = await got.post('http://service/endpoint', {
    body: JSON.stringify(requestBodyFixture),
  })

  t.equal(body, 'overwrite')
  scope.done()
})

test('reply can take a callback', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, (path, requestBody, callback) => callback(null, 'Hello World!'))

  const response = await got('http://example.com/', {
    encoding: null,
  })

  scope.done()
  t.type(response.body, Buffer)
  t.equal(response.body.toString('utf8'), 'Hello World!')
})

test('reply should send correct statusCode with array-notation and without body', async t => {
  t.plan(1)

  const expectedStatusCode = 202

  const scope = nock('http://example.com')
    .get('/')
    .reply((path, requestBody) => [expectedStatusCode])

  const { statusCode } = await got('http://example.com/')

  t.equal(statusCode, expectedStatusCode)
  scope.done()
})

test('reply takes a callback for status code', async t => {
  const expectedStatusCode = 202
  const responseBody = 'Hello, world!'
  const headers = {
    'X-Custom-Header': 'abcdef',
  }

  const scope = nock('http://example.com')
    .get('/')
    .reply((path, requestBody, cb) => {
      setTimeout(() => cb(null, [expectedStatusCode, responseBody, headers]), 1)
    })

  const response = await got('http://example.com/')

  t.equal(response.statusCode, expectedStatusCode, 'sends status code')
  t.deepEqual(response.headers, headers, 'sends headers')
  t.equal(response.body, responseBody, 'sends request body')
  scope.done()
})

test('reply should throw on error on the callback', t => {
  let dataCalled = false

  const scope = nock('http://example.com')
    .get('/')
    .reply(500, (path, requestBody, callback) =>
      callback(new Error('Database failed'))
    )

  // TODO When this request is converted to `got`, it causes the request not
  // to match.
  const req = http.request(
    {
      host: 'example.com',
      path: '/',
      port: 80,
    },
    res => {
      t.equal(res.statusCode, 500, 'Status code is 500')

      res.on('data', data => {
        dataCalled = true
        t.ok(data instanceof Buffer, 'data should be buffer')
        t.ok(
          data.toString().indexOf('Error: Database failed') === 0,
          'response should match'
        )
      })

      res.on('end', () => {
        t.ok(dataCalled, 'data handler was called')
        scope.done()
        t.end()
      })
    }
  )

  req.end()
})

test('match debugging works', async t => {
  const log = sinon.stub(debug, 'log')
  debug.enable('nock.interceptor')
  t.once('end', () => {
    sinon.restore()
    debug.disable('nock.interceptor')
  })

  nock('http://example.test')
    .post('/deep/link')
    .reply(200, 'Hello World!')

  const exampleBody = 'Hello yourself!'
  await got.post('http://example.test/deep/link', { body: exampleBody })

  t.ok(log.calledOnce)
  t.equal(
    JSON.parse(log.getCall(0).args[1]).href,
    'http://example.test/deep/link'
  )
  t.equal(JSON.parse(log.getCall(0).args[2]), exampleBody)
})

test('get gets mocked', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, 'Hello World!')

  const { statusCode, body } = await got('http://example.com/', {
    encoding: null,
  })

  t.equal(statusCode, 200)
  t.type(body, Buffer)
  t.equal(body.toString('utf8'), 'Hello World!')
  scope.done()
})

test('get gets mocked with relative base path', async t => {
  const scope = nock('http://example.com/abc')
    .get('/def')
    .reply(200, 'Hello World!')

  const { statusCode, body } = await got('http://example.com/abc/def', {
    encoding: null,
  })

  t.equal(statusCode, 200)
  t.type(body, Buffer)
  t.equal(body.toString('utf8'), 'Hello World!')
  scope.done()
})

test('post', async t => {
  const scope = nock('http://example.com')
    .post('/form')
    .reply(201, 'OK!')

  const { statusCode, body } = await got.post('http://example.com/form', {
    encoding: null,
  })

  t.equal(statusCode, 201)
  t.type(body, Buffer)
  t.equal(body.toString('utf8'), 'OK!')
  scope.done()
})

test('post with empty response body', async t => {
  const scope = nock('http://example.com')
    .post('/form')
    .reply(200)

  const { statusCode, body } = await got.post('http://example.com/form', {
    encoding: null,
  })

  t.equal(statusCode, 200)
  t.type(body, Buffer)
  t.equal(body.length, 0)
  scope.done()
})

test('post, lowercase', t => {
  let dataCalled = false

  const scope = nock('http://example.com')
    .post('/form')
    .reply(200, 'OK!')

  // Since this is testing a lowercase `method`, it's using the `http` module.
  const req = http.request(
    {
      host: 'example.com',
      method: 'post',
      path: '/form',
      port: 80,
    },
    res => {
      t.equal(res.statusCode, 200)
      res.on('end', () => {
        t.ok(dataCalled)
        scope.done()
        t.end()
      })
      res.on('data', data => {
        dataCalled = true
        t.ok(data instanceof Buffer, 'data should be buffer')
        t.equal(data.toString(), 'OK!', 'response should match')
      })
    }
  )

  req.end()
})

test('get with reply callback', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, () => 'OK!')

  const { body } = await got('http://example.com')
  t.equal(body, 'OK!')
  scope.done()
})

test('get to different subdomain with reply callback and filtering scope', async t => {
  // We scope for www.example.com but through scope filtering we will accept
  // any <subdomain>.example.com.
  const scope = nock('http://example.test', {
    filteringScope: scope => /^http:\/\/.*\.example/.test(scope),
  })
    .get('/')
    .reply(200, () => 'OK!')

  const { body } = await got('http://a.example.test')
  t.equal(body, 'OK!')
  scope.done()
})

test('get with reply callback returning object', async t => {
  const exampleResponse = { message: 'OK!' }

  const scope = nock('http://example.test')
    .get('/')
    .reply(200, () => exampleResponse)

  const { body } = await got('http://example.test')
  t.equal(body, JSON.stringify(exampleResponse))
  scope.done()
})

test('get with reply callback returning array with headers', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(() => [202, 'body', { 'x-key': 'value', 'x-key-2': 'value 2' }])

  const { headers, rawHeaders } = await got('http://example.test/')

  t.deepEqual(headers, {
    'x-key': 'value',
    'x-key-2': 'value 2',
  })
  t.deepEqual(rawHeaders, ['x-key', 'value', 'x-key-2', 'value 2'])
  scope.done()
})

// Skipped because https://github.com/nock/nock/issues/1222
test(
  'get with reply callback returning default statusCode without body',
  { skip: true },
  t => {
    nock('http://replyheaderland')
      .get('/')
      .reply((uri, requestBody) => [401])

    http.get(
      {
        host: 'replyheaderland',
        path: '/',
        port: 80,
      },
      res => {
        res.setEncoding('utf8')
        t.equal(res.statusCode, 200)
        res.on('data', data => {
          t.equal(data, '[401]')
          res.once('end', t.end.bind(t))
        })
      }
    )
  }
)

test('get with reply callback returning callback without headers', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(() => [401, 'This is a body'])

  await t.rejects(async () => got('http://example.com/'), {
    statusCode: 401,
    body: 'This is a body',
  })
  scope.done()
})

test('post with reply callback, uri, and request body', async t => {
  const input = 'key=val'

  const scope = nock('http://example.com')
    .post('/echo', input)
    .reply(200, (uri, body) => ['OK', uri, body].join(' '))

  const { body } = await got('http://example.com/echo', { body: input })
  t.equal(body, 'OK /echo key=val')
  scope.done()
})

test('post with regexp as spec', async t => {
  const input = 'key=val'

  const scope = nock('http://example.com')
    .post('/echo', /key=v.?l/g)
    .reply(200, (uri, body) => ['OK', uri, body].join(' '))

  const { body } = await got('http://example.com/echo', { body: input })

  t.equal(body, 'OK /echo key=val')
  scope.done()
})

test('post with function as spec', async t => {
  const scope = nock('http://example.com')
    .post('/echo', body => body === 'key=val')
    .reply(200, (uri, body) => ['OK', uri, body].join(' '))

  const { body } = await got('http://example.com/echo', { body: 'key=val' })

  t.equal(body, 'OK /echo key=val')
  scope.done()
})

test('post with chaining on call', async t => {
  const input = 'key=val'

  const scope = nock('http://example.com')
    .post('/echo', input)
    .reply(200, (uri, body) => ['OK', uri, body].join(' '))

  const { body } = await got('http://example.com/echo', { body: input })

  t.equal(body, 'OK /echo key=val')
  scope.done()
})

test('reply with callback and filtered path and body', async t => {
  let noPrematureExecution = false

  const scope = nock('http://example.com')
    .filteringPath(/.*/, '*')
    .filteringRequestBody(/.*/, '*')
    .post('*', '*')
    .reply(200, (uri, body) => {
      t.assert(noPrematureExecution)
      return ['OK', uri, body].join(' ')
    })

  noPrematureExecution = true
  const { body } = await got.post('http://example.com/original/path', {
    body: 'original=body',
  })

  t.equal(body, 'OK /original/path original=body')
  scope.done()
})

test('filteringPath with invalid argument throws expected', t => {
  t.throws(() => nock('http://example.test').filteringPath('abc123'), {
    message:
      'Invalid arguments: filtering path should be a function or a regular expression',
  })
  t.end()
})

test('filteringRequestBody with invalid argument throws expected', t => {
  t.throws(() => nock('http://example.test').filteringRequestBody('abc123'), {
    message:
      'Invalid arguments: filtering request body should be a function or a regular expression',
  })
  t.end()
})

test('isDone', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, 'Hello World!')

  t.notOk(scope.isDone(), 'not done when a request is outstanding')

  await got('http://example.com/')

  t.true(scope.isDone(), 'done after request is made')
  scope.done()
})

test('headers work', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, 'Hello World!', { 'X-My-Headers': 'My Header value' })

  const { headers } = await got('http://example.com/')

  t.equivalent(headers, { 'x-my-headers': 'My Header value' })
  scope.done()
})

test('reply headers work with function', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, () => 'ABC', { 'X-My-Headers': 'My custom header value' })

  const { headers } = await got('http://example.com/')

  t.equivalent(headers, { 'x-my-headers': 'My custom header value' })
  scope.done()
})

test('reply headers as function work', async t => {
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, 'boo!', {
      'X-My-Headers': (req, res, body) => body.toString(),
    })

  const { headers, rawHeaders } = await got('http://example.com/')

  t.equivalent(headers, { 'x-my-headers': 'boo!' })
  t.equivalent(rawHeaders, ['X-My-Headers', 'boo!'])
  scope.done()
})

test('reply headers as function are evaluated only once per request', async t => {
  let counter = 0
  const scope = nock('http://example.com')
    .get('/')
    .reply(200, 'boo!', {
      'X-My-Headers': (req, res, body) => {
        ++counter
        return body.toString()
      },
    })

  const { headers, rawHeaders } = await got('http://example.com/')

  t.equivalent(headers, { 'x-my-headers': 'boo!' })
  t.equivalent(rawHeaders, ['X-My-Headers', 'boo!'])
  scope.done()

  t.equal(counter, 1)
})

test('reply headers as function are evaluated on each request', async t => {
  let counter = 0
  const scope = nock('http://example.com')
    .get('/')
    .times(2)
    .reply(200, 'boo!', {
      'X-My-Headers': (req, res, body) => `${++counter}`,
    })

  const { headers, rawHeaders } = await got('http://example.com/')
  t.equivalent(headers, { 'x-my-headers': '1' })
  t.equivalent(rawHeaders, ['X-My-Headers', '1'])

  t.equal(counter, 1)

  const { headers: headers2, rawHeaders: rawHeaders2 } = await got(
    'http://example.com/'
  )
  t.equivalent(headers2, { 'x-my-headers': '2' })
  t.equivalent(rawHeaders2, ['X-My-Headers', '2'])

  t.equal(counter, 2)

  scope.done()
})

test('match headers', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .matchHeader('x-my-headers', 'My custom Header value')
    .reply(200, 'Hello World!')

  const { statusCode, body } = await got('http://example.test/', {
    headers: { 'X-My-Headers': 'My custom Header value' },
  })

  t.equal(statusCode, 200)
  t.equal(body, 'Hello World!')
  scope.done()
})

test('multiple match headers', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .matchHeader('x-my-headers', 'My custom Header value')
    .reply(200, 'Hello World!')
    .get('/')
    .matchHeader('x-my-headers', 'other value')
    .reply(200, 'Hello World other value!')

  const response1 = await got('http://example.test/', {
    headers: { 'X-My-Headers': 'other value' },
  })

  t.equal(response1.statusCode, 200)
  t.equal(response1.body, 'Hello World other value!')

  const response2 = await got('http://example.test/', {
    headers: { 'X-My-Headers': 'My custom Header value' },
  })

  t.equal(response2.statusCode, 200)
  t.equal(response2.body, 'Hello World!')

  scope.done()
})

test('match headers with regexp', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .matchHeader('x-my-headers', /My He.d.r [0-9.]+/)
    .reply(200, 'Hello World!')

  const { statusCode, body } = await got('http://example.test/', {
    headers: { 'X-My-Headers': 'My Header 1.0' },
  })

  t.equal(statusCode, 200)
  t.equal(body, 'Hello World!')
  scope.done()
})

test('match headers on number with regexp', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .matchHeader('x-my-headers', /\d+/)
    .reply(200, 'Hello World!')

  const { statusCode, body } = await got('http://example.test/', {
    headers: { 'X-My-Headers': 123 },
  })

  t.equal(statusCode, 200)
  t.equal(body, 'Hello World!')
  scope.done()
})

test('match headers with function', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .matchHeader('x-my-headers', val => val > 123)
    .reply(200, 'Hello World!')

  const { statusCode, body } = await got('http://example.test/', {
    headers: { 'X-My-Headers': 456 },
  })

  t.equal(statusCode, 200)
  t.equal(body, 'Hello World!')
  scope.done()
})

test('match all headers', async t => {
  const scope = nock('http://example.test')
    .matchHeader('accept', 'application/json')
    .get('/one')
    .reply(200, { hello: 'world' })
    .get('/two')
    .reply(200, { a: 1, b: 2, c: 3 })

  const response1 = await got('http://example.test/one', {
    headers: { Accept: 'application/json' },
  })
  t.equal(response1.statusCode, 200)
  t.equal(response1.body, '{"hello":"world"}')

  const response2 = await got('http://example.test/two', {
    headers: { Accept: 'application/json' },
  })
  t.equal(response2.statusCode, 200)
  t.equal(response2.body, '{"a":1,"b":2,"c":3}')

  scope.done()
})

test('header manipulation', t => {
  // This test seems to depend on behavior of the `http` module.
  const scope = nock('http://example.com')
    .get('/accounts')
    .reply(200, { accounts: [{ id: 1, name: 'Joe Blow' }] })

  const req = http.get({ host: 'example.com', path: '/accounts' }, res => {
    res.on('end', () => {
      scope.done()
      t.end()
    })
    // Streams start in 'paused' mode and must be started.
    // See https://nodejs.org/api/stream.html#stream_class_stream_readable
    res.resume()
  })

  req.setHeader('X-Custom-Header', 'My Value')
  t.equal(
    req.getHeader('X-Custom-Header'),
    'My Value',
    'Custom header was not set'
  )

  req.removeHeader('X-Custom-Header')
  t.notOk(req.getHeader('X-Custom-Header'), 'Custom header was not removed')

  req.end()
})

test('head', async t => {
  const scope = nock('http://example.test')
    .head('/')
    .reply(201, 'OK!')

  const { statusCode } = await got.head('http://example.test/')

  t.equal(statusCode, 201)
  scope.done()
})

test('body data is differentiating', async t => {
  const scope = nock('http://example.test')
    .post('/', 'abc')
    .reply(200, 'Hey 1')
    .post('/', 'def')
    .reply(200, 'Hey 2')

  const response1 = await got('http://example.test/', { body: 'abc' })
  t.equal(response1.statusCode, 200)
  t.equal(response1.body, 'Hey 1')

  const response2 = await got('http://example.test/', { body: 'def' })
  t.equal(response2.statusCode, 200)
  t.equal(response2.body, 'Hey 2')

  scope.done()
})

test('chaining', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')
    .post('/form')
    .reply(201, 'OK!')

  const response1 = await got.post('http://example.test/form')
  t.equal(response1.statusCode, 201)
  t.equal(response1.body, 'OK!')

  const response2 = await got('http://example.test/')
  t.equal(response2.statusCode, 200)
  t.equal(response2.body, 'Hello World!')

  scope.done()
})

test('encoding', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  const { body } = await got('http://example.test/', { encoding: 'base64' })

  t.type(body, 'string')
  t.equal(body, 'SGVsbG8gV29ybGQh', 'response should match base64 encoding')

  scope.done()
})

test('reply with file', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .replyWithFile(200, textFile)

  const { statusCode, body } = await got('http://example.test/')

  t.equal(statusCode, 200)
  t.equal(body, 'Hello from the file!')

  scope.done()
})

test('reply with file with headers', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .replyWithFile(200, binaryFile, {
      'content-encoding': 'gzip',
    })

  const { statusCode, body } = await got('http://example.test/')

  t.equal(statusCode, 200)
  t.equal(body.length, 20)
  scope.done()
})

test('reply with JSON', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(200, { hello: 'world' })

  const { statusCode, headers, body } = await got('http://example.test/')

  t.equal(statusCode, 200)
  t.type(headers.date, 'undefined')
  t.type(headers['content-length'], 'undefined')
  t.equal(headers['content-type'], 'application/json')
  t.equal(body, '{"hello":"world"}', 'response should match')
  scope.done()
})

test('reply with content-length header', async t => {
  const scope = nock('http://example.test')
    .replyContentLength()
    .get('/')
    .reply(200, { hello: 'world' })

  const { headers } = await got('http://example.test/')

  t.equal(headers['content-length'], 17)
  scope.done()
})

test('reply with explicit date header', async t => {
  const date = new Date()

  const scope = nock('http://example.test')
    .replyDate(date)
    .get('/')
    .reply(200, { hello: 'world' })

  const { headers } = await got('http://example.test/')

  t.equal(headers.date, date.toUTCString())
  scope.done()
})

// async / got version is returning "not ok test unfinished".
// https://github.com/nock/nock/issues/1305#issuecomment-451701657
test('reply with implicit date header', t => {
  const clock = lolex.install()
  const date = new Date()

  const scope = nock('http://example.test')
    .replyDate()
    .get('/')
    .reply(200)

  mikealRequest.get('http://example.test', (err, resp) => {
    clock.uninstall()

    if (err) {
      throw err
    }

    t.equal(resp.headers.date, date.toUTCString())
    scope.done()

    t.end()
  })
})

test('filter path with function', async t => {
  const scope = nock('http://example.test')
    .filteringPath(path => '/?a=2&b=1')
    .get('/?a=2&b=1')
    .reply(200, 'Hello World!')

  const { statusCode } = await got('http://example.test/', {
    query: { a: '1', b: '2' },
  })

  t.equal(statusCode, 200)
  scope.done()
})

test('filter path with regexp', async t => {
  const scope = nock('http://example.test')
    .filteringPath(/\d/g, '3')
    .get('/?a=3&b=3')
    .reply(200, 'Hello World!')

  const { statusCode } = await got('http://example.test/', {
    query: { a: '1', b: '2' },
  })

  t.equal(statusCode, 200)
  scope.done()
})

test('filter body with function', async t => {
  let filteringRequestBodyCounter = 0

  const scope = nock('http://example.test')
    .filteringRequestBody(body => {
      ++filteringRequestBodyCounter
      t.equal(body, 'mamma mia')
      return 'mamma tua'
    })
    .post('/', 'mamma tua')
    .reply(200, 'Hello World!')

  const { statusCode } = await got('http://example.test/', {
    body: 'mamma mia',
  })

  t.equal(statusCode, 200)
  scope.done()
  t.equal(filteringRequestBodyCounter, 1)
})

test('filter body with regexp', async t => {
  const scope = nock('http://example.test')
    .filteringRequestBody(/mia/, 'nostra')
    .post('/', 'mamma nostra')
    .reply(200, 'Hello World!')

  const { statusCode } = await got('http://example.test/', {
    body: 'mamma mia',
  })

  t.equal(statusCode, 200)
  scope.done()
})

// TODO Convert to async / got.
test('abort request', t => {
  const scope = nock('http://example.test')
    .get('/hey')
    .reply(200, 'nobody')

  const req = http.request({
    host: 'example.test',
    path: '/hey',
  })

  req.on('response', res => {
    res.on('close', err => {
      t.equal(err.code, 'aborted')
      scope.done()
    })

    res.on('end', () => t.fail('this should never execute'))

    req.once('error', err => {
      t.equal(err.code, 'ECONNRESET')
      t.end()
    })

    req.abort()
  })

  req.end()
})

// TODO Convert to async / got.
test('pause response before data', t => {
  const scope = nock('http://example.test')
    .get('/pauser')
    .reply(200, 'nobody')

  const req = http.request({
    host: 'example.test',
    path: '/pauser',
  })

  req.on('response', res => {
    res.pause()

    let waited = false
    setTimeout(() => {
      waited = true
      res.resume()
    }, 500)

    res.on('data', data => t.true(waited))

    res.on('end', () => {
      scope.done()
      t.end()
    })
  })

  req.end()
})

test('chaining API', async t => {
  const scope = nock('http://example.test')
    .get('/one')
    .reply(200, 'first one')
    .get('/two')
    .reply(200, 'second one')

  const response1 = await got('http://example.test/one')

  t.equal(response1.statusCode, 200)
  t.equal(response1.body, 'first one')

  const response2 = await got('http://example.test/two')

  t.equal(response2.statusCode, 200)
  t.equal(response2.body, 'second one')

  scope.done()
})

test('same URI', async t => {
  const scope = nock('http://example.test')
    .get('/abc')
    .reply(200, 'first one')
    .get('/abc')
    .reply(201, 'second one')

  const response1 = await got('http://example.test/abc')

  t.equal(response1.statusCode, 200)
  t.equal(response1.body, 'first one')

  const response2 = await got('http://example.test/abc')

  t.equal(response2.statusCode, 201)
  t.equal(response2.body, 'second one')

  scope.done()
})

// TODO Should this test be kept?
test('can use hostname instead of host', t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  const req = http.request(
    {
      hostname: 'example.test',
      path: '/',
    },
    function(res) {
      t.equal(res.statusCode, 200)
      res.on('end', function() {
        scope.done()
        t.end()
      })
      // Streams start in 'paused' mode and must be started.
      // See https://nodejs.org/api/stream.html#stream_class_stream_readable
      res.resume()
    }
  )

  req.end()
})

// TODO convert to async / got.
test('hostname is case insensitive', t => {
  const scope = nock('http://example.test')
    .get('/path')
    .reply(200, 'hey')

  const options = {
    hostname: 'example.test',
    path: '/path',
    method: 'GET',
  }

  const req = http.request(options, function(res) {
    scope.done()
    t.end()
  })

  req.end()
})

test('can take a port', async t => {
  const scope = nock('http://example.test:3333')
    .get('/')
    .reply(200, 'Hello World!')

  const { statusCode } = await got('http://example.test:3333/')

  t.equal(statusCode, 200)
  scope.done()
})

test('can use https', async t => {
  const scope = nock('https://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  const { statusCode, body } = await got('https://example.test/', {
    encoding: null,
  })

  t.equal(statusCode, 200)
  t.type(body, Buffer)
  t.equal(body.toString(), 'Hello World!')
  scope.done()
})

// TODO convert to got / async.
test('emits error if https route is missing', t => {
  nock('https://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  const req = https.request(
    {
      host: 'example.test',
      path: '/abcdef892932',
    },
    function(res) {
      throw new Error('should not come here!')
    }
  )

  req.end()

  // This listener is intentionally after the end call so make sure that
  // listeners added after the end will catch the error
  req.on('error', function(err) {
    t.equal(
      err.message.trim(),
      `Nock: No match for request ${JSON.stringify(
        { method: 'GET', url: 'https://example.test/abcdef892932' },
        null,
        2
      )}`
    )
    t.end()
  })
})

// TODO convert to got / async.
test('emits error if https route is missing', t => {
  nock('https://example.test:123')
    .get('/')
    .reply(200, 'Hello World!')

  const req = https.request(
    {
      host: 'example.test',
      port: 123,
      path: '/dsadsads',
    },
    function(res) {
      throw new Error('should not come here!')
    }
  )

  req.end()

  // This listener is intentionally after the end call so make sure that
  // listeners added after the end will catch the error
  req.on('error', function(err) {
    t.equal(
      err.message.trim(),
      `Nock: No match for request ${JSON.stringify(
        { method: 'GET', url: 'https://example.test:123/dsadsads' },
        null,
        2
      )}`
    )
    t.end()
  })
})

test('scopes are independent', async t => {
  const scope1 = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')
  const scope2 = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  await got('http://example.test/')

  t.true(scope1.isDone())
  t.false(scope2.isDone())

  nock.cleanAll()
})

test('two scopes with the same request are consumed', async t => {
  const scope1 = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  const scope2 = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  await got('http://example.test/')
  await got('http://example.test/')

  scope1.done()
  scope2.done()
})

test('JSON encoded replies set the content-type header', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(200, {
      A: 'b',
    })

  t.equal(
    (await got('http://example.test/')).headers['content-type'],
    'application/json'
  )

  scope.done()
})

test('JSON encoded replies does not overwrite existing content-type header', async t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(
      200,
      {
        A: 'b',
      },
      {
        'Content-Type': 'unicorns',
      }
    )

  t.equal(
    (await got('http://example.test/')).headers['content-type'],
    'unicorns'
  )

  scope.done()
})

test("blank response doesn't have content-type application/json attached to it", async t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(200)

  t.equal(
    (await got('http://example.test/')).headers['content-type'],
    undefined
  )

  scope.done()
})

test('unencodable object throws the expected error', t => {
  const unencodableObject = {
    toJSON() {
      throw Error('bad!')
    },
  }

  t.throws(
    () =>
      nock('http://localhost')
        .get('/')
        .reply(200, unencodableObject),
    {
      message: 'Error encoding response body into JSON',
    }
  )

  t.end()
})

test('clean all works', t => {
  nock('http://example.test')
    .get('/nonexistent')
    .reply(200)

  http.get({ host: 'example.test', path: '/nonexistent' }, function(res) {
    t.assert(res.statusCode === 200, 'should mock before cleanup')

    nock.cleanAll()

    http
      .get({ host: 'example.test', path: '/nonexistent' }, function(res) {
        res.destroy()
        t.assert(res.statusCode !== 200, 'should clean up properly')
        t.end()
      })
      .on('error', function(err) {
        t.end()
      })
  })
})

test('cleanAll should remove pending mocks from all scopes', t => {
  const scope1 = nock('http://example.test')
    .get('/somepath')
    .reply(200, 'hey')
  t.deepEqual(scope1.pendingMocks(), ['GET http://example.test:80/somepath'])
  const scope2 = nock('http://example.test')
    .get('/somepath')
    .reply(200, 'hey')
  t.deepEqual(scope2.pendingMocks(), ['GET http://example.test:80/somepath'])

  nock.cleanAll()

  t.deepEqual(scope1.pendingMocks(), [])
  t.deepEqual(scope2.pendingMocks(), [])
  t.end()
})

test('is done works', t => {
  nock('http://example.test')
    .get('/nonexistent')
    .reply(200)

  t.ok(!nock.isDone())

  http.get({ host: 'example.test', path: '/nonexistent' }, function(res) {
    t.assert(res.statusCode === 200, 'should mock before cleanup')
    t.ok(nock.isDone())
    t.end()
  })
})

test('pending mocks works', t => {
  nock('http://example.test')
    .get('/nonexistent')
    .reply(200)

  t.deepEqual(nock.pendingMocks(), ['GET http://example.test:80/nonexistent'])

  http.get({ host: 'example.test', path: '/nonexistent' }, function(res) {
    t.assert(res.statusCode === 200, 'should mock before cleanup')
    t.deepEqual(nock.pendingMocks(), [])
    t.end()
  })
})

test('activeMocks returns incomplete mocks', t => {
  nock.cleanAll()
  nock('http://example.test')
    .get('/incomplete')
    .reply(200)

  t.deepEqual(nock.activeMocks(), ['GET http://example.test:80/incomplete'])
  t.end()
})

test("activeMocks doesn't return completed mocks", t => {
  nock.cleanAll()
  nock('http://example.test')
    .get('/complete-me')
    .reply(200)

  http.get({ host: 'example.test', path: '/complete-me' }, function(res) {
    t.deepEqual(nock.activeMocks(), [])
    t.end()
  })
})

test('username and password works', t => {
  const scope = nock('http://example.test')
    .get('/')
    .reply(200, 'Welcome, username')

  http
    .request(
      {
        hostname: 'example.test',
        auth: 'username:password',
        path: '/',
      },
      function(res) {
        scope.done()
        t.end()
      }
    )
    .end()
})

test('works with mikeal/request and username and password', t => {
  const scope = nock('http://example.test')
    .get('/abc')
    .reply(200, 'Welcome, username')

  mikealRequest(
    { uri: 'http://username:password@example.test/abc', log: true },
    function(err, res, body) {
      t.ok(!err, 'error')
      t.ok(scope.isDone())
      t.equal(body, 'Welcome, username')
      t.end()
    }
  )
})

test('different ports work works', t => {
  const scope = nock('http://example.test:8081')
    .get('/pathhh')
    .reply(200, 'Welcome, username')

  http
    .request(
      {
        hostname: 'example.test',
        port: 8081,
        path: '/pathhh',
      },
      function(res) {
        scope.done()
        t.end()
      }
    )
    .end()
})

test('different ports work work with Mikeal request', t => {
  const scope = nock('http://example.test:8082')
    .get('/pathhh')
    .reply(200, 'Welcome to Mikeal Request!')

  mikealRequest.get('http://example.test:8082/pathhh', function(
    err,
    res,
    body
  ) {
    t.ok(!err, 'no error')
    t.equal(body, 'Welcome to Mikeal Request!')
    t.ok(scope.isDone())
    t.end()
  })
})

test('explicitly specifiying port 80 works', t => {
  const scope = nock('http://example.test:80')
    .get('/pathhh')
    .reply(200, 'Welcome, username')

  http
    .request(
      {
        hostname: 'example.test',
        port: 80,
        path: '/pathhh',
      },
      function(res) {
        scope.done()
        t.end()
      }
    )
    .end()
})

test('post with object', t => {
  const scope = nock('http://example.test')
    .post('/claim', { some_data: 'something' })
    .reply(200)

  http
    .request(
      {
        hostname: 'example.test',
        port: 80,
        method: 'POST',
        path: '/claim',
      },
      function(res) {
        scope.done()
        t.end()
      }
    )
    .end('{"some_data":"something"}')
})

test('accept string as request target', t => {
  let dataCalled = false
  const scope = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello World!')

  http.get('http://example.test', function(res) {
    t.equal(res.statusCode, 200)

    res.on('data', function(data) {
      dataCalled = true
      t.ok(data instanceof Buffer, 'data should be buffer')
      t.equal(data.toString(), 'Hello World!', 'response should match')
    })

    res.on('end', function() {
      t.ok(dataCalled)
      scope.done()
      t.end()
    })
  })
})

if (url.URL) {
  test('accept URL as request target', t => {
    let dataCalled = false
    const scope = nock('http://example.test')
      .get('/')
      .reply(200, 'Hello World!')

    http.get(new url.URL('http://example.test'), function(res) {
      t.equal(res.statusCode, 200)

      res.on('data', function(data) {
        dataCalled = true
        t.ok(data instanceof Buffer, 'data should be buffer')
        t.equal(data.toString(), 'Hello World!', 'response should match')
      })

      res.on('end', function() {
        t.ok(dataCalled)
        scope.done()
        t.end()
      })
    })
  })
}

test('request has path', t => {
  const scope = nock('http://example.test')
    .get('/the/path/to/infinity')
    .reply(200)

  const req = http.request(
    {
      hostname: 'example.test',
      port: 80,
      method: 'GET',
      path: '/the/path/to/infinity',
    },
    function(res) {
      scope.done()
      t.equal(
        req.path,
        '/the/path/to/infinity',
        'should have req.path set to /the/path/to/infinity'
      )
      t.end()
    }
  )
  req.end()
})

test('(re-)activate after restore', t => {
  t.plan(7)

  const server = http.createServer((request, response) => {
    t.pass('server received a request')

    switch (url.parse(request.url).pathname) {
      case '/':
        response.writeHead(200)
        response.write('server served a response')
        break
    }

    response.end()
  })

  server.listen(() => {
    const scope = nock(`http://localhost:${server.address().port}`)
      .get('/')
      .reply(304, 'served from our mock')

    nock.restore()
    t.false(nock.isActive())

    http.get(`http://localhost:${server.address().port}`, function(res) {
      res.resume()

      t.is(200, res.statusCode)

      res.on('end', function() {
        t.ok(!scope.isDone())

        nock.activate()
        t.true(nock.isActive())
        http.get(`http://localhost:${server.address().port}`, function(res) {
          res.resume()

          t.is(304, res.statusCode)

          res.on('end', function() {
            t.ok(scope.isDone())

            server.close(t.end)
          })
        })
      })
    })
  })
})

test('allow unordered body with json encoding', t => {
  const scope = nock('http://example.test')
    .post('/like-wtf', {
      foo: 'bar',
      bar: 'foo',
    })
    .reply(200, 'Heyyyy!')

  mikealRequest(
    {
      uri: 'http://example.test/like-wtf',
      method: 'POST',
      json: {
        bar: 'foo',
        foo: 'bar',
      },
    },
    function(e, r, body) {
      t.equal(body, 'Heyyyy!')
      scope.done()
      t.end()
    }
  )
})

test('allow unordered body with form encoding', t => {
  const scope = nock('http://example.test')
    .post('/like-wtf', {
      foo: 'bar',
      bar: 'foo',
    })
    .reply(200, 'Heyyyy!')

  mikealRequest(
    {
      uri: 'http://example.test/like-wtf',
      method: 'POST',
      form: {
        bar: 'foo',
        foo: 'bar',
      },
    },
    function(e, r, body) {
      t.equal(body, 'Heyyyy!')
      scope.done()
      t.end()
    }
  )
})

test('allow string json spec', t => {
  const bodyObject = { bar: 'foo', foo: 'bar' }

  const scope = nock('http://example.test')
    .post('/like-wtf', JSON.stringify(bodyObject))
    .reply(200, 'Heyyyy!')

  mikealRequest(
    {
      uri: 'http://example.test/like-wtf',
      method: 'POST',
      json: {
        bar: 'foo',
        foo: 'bar',
      },
    },
    function(e, r, body) {
      t.equal(body, 'Heyyyy!')
      scope.done()
      t.end()
    }
  )
})

test('has a req property on the response', t => {
  const scope = nock('http://example.test')
    .get('/like-wtf')
    .reply(200)
  const req = http.request('http://example.test/like-wtf', function(res) {
    res.on('end', function() {
      t.ok(res.req, "req property doesn't exist")
      scope.done()
      t.end()
    })
    // Streams start in 'paused' mode and must be started.
    // See https://nodejs.org/api/stream.html#stream_class_stream_readable
    res.resume()
  })
  req.end()
})

test('when net connect is disabled, throws the expected error ', async t => {
  nock.cleanAll()
  nock.disableNetConnect()
  t.once('end', () => nock.enableNetConnect())

  try {
    await got('http://example.test')
    t.fail('Expected to throw')
  } catch (err) {
    t.type(err, 'Error')
    t.equal(err.message, 'Nock: Disallowed net connect for "example.test:80/"')
    t.equal(err.code, 'ENETUNREACH')
    t.ok(err.stack)
  }
})

test('enable real HTTP request only for specified domain, via string', t => {
  t.plan(1)

  const server = http.createServer((request, response) => {
    t.pass('server received a request')
    response.writeHead(200)
    response.end()
    t.end()
  })
  t.once('end', () => server.close())

  nock.enableNetConnect('localhost')
  t.once('end', () => nock.enableNetConnect())

  server.listen(() =>
    mikealRequest(`http://localhost:${server.address().port}/`)
  )
})

test('disallow request for other domains, via string', t => {
  nock.enableNetConnect('localhost')
  t.once('end', () => nock.enableNetConnect())

  http
    .get('http://www.amazon.com', function(res) {
      throw 'should not deliver this request'
    })
    .on('error', function(err) {
      t.equal(
        err.message,
        'Nock: Disallowed net connect for "www.amazon.com:80/"'
      )
      t.end()
    })
})

test('enable real HTTP request only for specified domain, via regexp', t => {
  t.plan(1)

  const server = http.createServer((request, response) => {
    t.pass('server received a request')
    response.writeHead(200)
    response.end()
    t.end()
  })
  t.once('end', () => server.close())

  nock.enableNetConnect(/ocalhos/)
  t.once('end', () => nock.enableNetConnect())

  server.listen(() =>
    mikealRequest(`http://localhost:${server.address().port}/`)
  )
})

test('disallow request for other domains, via regexp', t => {
  nock.enableNetConnect(/ocalhos/)
  t.once('end', () => nock.enableNetConnect())

  http
    .get('http://www.amazon.com', function(res) {
      throw 'should not deliver this request'
    })
    .on('error', function(err) {
      t.equal(
        err.message,
        'Nock: Disallowed net connect for "www.amazon.com:80/"'
      )
      t.end()
    })
})

test('repeating once', t => {
  nock.disableNetConnect()

  nock('http://example.test')
    .get('/')
    .once()
    .reply(200, 'Hello World!')

  http.get('http://example.test', function(res) {
    t.equal(200, res.statusCode, 'first request')
    t.end()
  })

  nock.cleanAll()

  nock.enableNetConnect()
})

test('repeating twice', t => {
  nock.disableNetConnect()

  nock('http://example.test')
    .get('/')
    .twice()
    .reply(200, 'Hello World!')

  async.each(
    [1, 2],
    function(_, cb) {
      http.get('http://example.test', function(res) {
        t.equal(200, res.statusCode)
        cb()
      })
    },
    t.end.bind(t)
  )
})

test('repeating thrice', t => {
  nock.disableNetConnect()

  nock('http://example.test')
    .get('/')
    .thrice()
    .reply(200, 'Hello World!')

  async.each(
    [1, 2, 3],
    function(_, cb) {
      http.get('http://example.test', function(res) {
        t.equal(200, res.statusCode)
        cb()
      })
    },
    t.end.bind(t)
  )
})

test('repeating response 4 times', t => {
  nock.disableNetConnect()

  nock('http://example.test')
    .get('/')
    .times(4)
    .reply(200, 'Hello World!')

  async.each(
    [1, 2, 3, 4],
    function(_, cb) {
      http.get('http://example.test', function(res) {
        t.equal(200, res.statusCode, 'first request')
        cb()
      })
    },
    t.end.bind(t)
  )
})

test('superagent works', t => {
  const responseText = 'Yay superagent!'
  const headers = { 'Content-Type': 'text/plain' }
  nock('http://example.test')
    .get('/somepath')
    .reply(200, responseText, headers)

  superagent.get('http://example.test/somepath').end(function(err, res) {
    t.equal(res.text, responseText)
    t.end()
  })
})

test('superagent works with query string', t => {
  const responseText = 'Yay superagentzzz'
  const headers = { 'Content-Type': 'text/plain' }
  nock('http://example.test')
    .get('/somepath?a=b')
    .reply(200, responseText, headers)

  superagent.get('http://example.test/somepath?a=b').end(function(err, res) {
    t.equal(res.text, responseText)
    t.end()
  })
})

test('superagent posts', t => {
  nock('http://example.test')
    .post('/somepath?b=c')
    .reply(204)

  superagent
    .post('http://example.test/somepath?b=c')
    .send('some data')
    .end(function(err, res) {
      t.equal(res.status, 204)
      t.end()
    })
})

test('response is an http.IncomingMessage instance', t => {
  const responseText = 'incoming message!'
  nock('http://example.test')
    .get('/somepath')
    .reply(200, responseText)

  http
    .request(
      {
        host: 'example.test',
        path: '/somepath',
      },
      function(res) {
        res.resume()
        t.true(res instanceof http.IncomingMessage)
        t.end()
      }
    )
    .end()
})

test('resetting nock catastrophically while a request is in progress is handled gracefully', async t => {
  // While invoking cleanAll() from a nock request handler isn't very
  // realistic, it's possible that user code under test could crash, causing
  // before or after hooks to fire, which invoke `nock.cleanAll()`. A little
  // extreme, though if this does happen, we may as well be graceful about it.
  function somethingBad() {
    nock.cleanAll()
  }

  const scope = nock('http://example.test')
    .get('/somepath')
    .reply(200, (uri, requestBody) => {
      somethingBad()
      return 'hi'
    })

  const { body } = await got('http://example.test/somepath')

  t.equal(body, 'hi')
  scope.done()
})

test('write callback called', t => {
  const scope = nock('http://filterboddiezregexp.com')
    .filteringRequestBody(/mia/, 'nostra')
    .post('/', 'mamma nostra')
    .reply(200, 'Hello World!')

  let callbackCalled = false
  const req = http.request(
    {
      host: 'filterboddiezregexp.com',
      method: 'POST',
      path: '/',
      port: 80,
    },
    function(res) {
      t.equal(callbackCalled, true)
      t.equal(res.statusCode, 200)
      res.on('end', function() {
        scope.done()
        t.end()
      })
      // Streams start in 'paused' mode and must be started.
      // See https://nodejs.org/api/stream.html#stream_class_stream_readable
      res.resume()
    }
  )

  req.write('mamma mia', null, function() {
    callbackCalled = true
    req.end()
  })
})

test('end callback called', t => {
  const scope = nock('http://example.test')
    .filteringRequestBody(/mia/, 'nostra')
    .post('/', 'mamma nostra')
    .reply(200, 'Hello World!')

  let callbackCalled = false
  const req = http.request(
    {
      host: 'example.test',
      method: 'POST',
      path: '/',
      port: 80,
    },
    function(res) {
      t.equal(callbackCalled, true)
      t.equal(res.statusCode, 200)
      res.on('end', function() {
        scope.done()
        t.end()
      })
      // Streams start in 'paused' mode and must be started.
      // See https://nodejs.org/api/stream.html#stream_class_stream_readable
      res.resume()
    }
  )

  req.end('mamma mia', null, function() {
    callbackCalled = true
  })
})

test('finish event fired before end event (bug-139)', t => {
  const scope = nock('http://example.test')
    .filteringRequestBody(/mia/, 'nostra')
    .post('/', 'mamma nostra')
    .reply(200, 'Hello World!')

  let finishCalled = false
  const req = http.request(
    {
      host: 'example.test',
      method: 'POST',
      path: '/',
      port: 80,
    },
    function(res) {
      t.equal(finishCalled, true)
      t.equal(res.statusCode, 200)
      res.on('end', function() {
        scope.done()
        t.end()
      })
      // Streams start in 'paused' mode and must be started.
      // See https://nodejs.org/api/stream.html#stream_class_stream_readable
      res.resume()
    }
  )

  req.on('finish', function() {
    finishCalled = true
  })

  req.end('mamma mia')
})

test('sending binary and receiving JSON should work ', t => {
  const scope = nock('http://example.test')
    .filteringRequestBody(/.*/, '*')
    .post('/some/path', '*')
    .reply(
      201,
      { foo: '61' },
      {
        'Content-Type': 'application/json',
      }
    )

  mikealRequest(
    {
      method: 'POST',
      uri: 'http://example.test/some/path',
      body: Buffer.from('ffd8ffe000104a46494600010101006000600000ff', 'hex'),
      headers: { Accept: 'application/json', 'Content-Length': 23861 },
    },
    function(err, res, body) {
      scope.done()

      t.equal(res.statusCode, 201)
      t.equal(body.length, 12)

      let json
      try {
        json = JSON.parse(body)
      } catch (e) {
        json = {}
      }

      t.equal(json.foo, '61')
      t.end()
    }
  )
})

// https://github.com/nock/nock/issues/146
test('resume() is automatically invoked when the response is drained', t => {
  const replyLength = 1024 * 1024
  const replyBuffer = Buffer.from(new Array(replyLength + 1).join('.'))
  t.equal(replyBuffer.length, replyLength)

  nock('http://example.test')
    .get('/abc')
    .reply(200, replyBuffer)

  needle.get('http://example.test/abc', function(err, res, buffer) {
    t.notOk(err)
    t.ok(res)
    t.ok(buffer)
    t.same(buffer, replyBuffer)
    t.end()
  })
})

test('handles get with restify client', t => {
  const scope = nock('https://example.test')
    .get('/get')
    .reply(200, 'get')

  const client = restify.createClient({
    url: 'https://example.test',
  })

  client.get('/get', function(err, req, res) {
    req.on('result', function(err, res) {
      res.body = ''
      res.setEncoding('utf8')
      res.on('data', function(chunk) {
        res.body += chunk
      })

      res.on('end', function() {
        t.equal(res.body, 'get')
        t.end()
        scope.done()
      })
    })
  })
})

test('handles post with restify client', t => {
  const scope = nock('https://example.test')
    .post('/post', 'hello world')
    .reply(200, 'post')

  const client = restify.createClient({
    url: 'https://example.test',
  })

  client.post('/post', function(err, req, res) {
    req.on('result', function(err, res) {
      res.body = ''
      res.setEncoding('utf8')
      res.on('data', function(chunk) {
        res.body += chunk
      })

      res.on('end', function() {
        t.equal(res.body, 'post')
        t.end()
        scope.done()
      })
    })

    req.write('hello world')
    req.end()
  })
})

test('handles get with restify JsonClient', t => {
  const scope = nock('https://example.test')
    .get('/get')
    .reply(200, { get: 'ok' })

  const client = restify.createJsonClient({
    url: 'https://example.test',
  })

  client.get('/get', function(err, req, res, obj) {
    t.equal(obj.get, 'ok')
    t.end()
    scope.done()
  })
})

test('handles post with restify JsonClient', t => {
  const scope = nock('https://example.test')
    .post('/post', { username: 'banana' })
    .reply(200, { post: 'ok' })

  const client = restify.createJsonClient({
    url: 'https://example.test',
  })

  client.post('/post', { username: 'banana' }, function(err, req, res, obj) {
    t.equal(obj.post, 'ok')
    t.end()
    scope.done()
  })
})

test('handles 404 with restify JsonClient', t => {
  const scope = nock('https://example.test')
    .put('/404')
    .reply(404)

  const client = restify.createJsonClient({
    url: 'https://example.test',
  })

  client.put('/404', function(err, req, res, obj) {
    t.equal(res.statusCode, 404)
    t.end()
    scope.done()
  })
})

test('handles 500 with restify JsonClient', t => {
  const scope = nock('https://example.test')
    .delete('/500')
    .reply(500)

  const client = restify.createJsonClient({
    url: 'https://example.test',
  })

  client.del('/500', function(err, req, res, obj) {
    t.equal(res.statusCode, 500)
    t.end()
    scope.done()
  })
})

test('test request timeout option', t => {
  nock('http://example.com')
    .get('/test')
    .reply(200, JSON.stringify({ foo: 'bar' }))

  const options = {
    url: 'http://example.com/test',
    method: 'GET',
    timeout: 2000,
  }

  mikealRequest(options, function(err, res, body) {
    t.strictEqual(err, null)
    t.equal(body, '{"foo":"bar"}')
    t.end()
  })
})

test('done fails when specified request header is missing', t => {
  nock('http://example.test', {
    reqheaders: {
      'X-App-Token': 'apptoken',
      'X-Auth-Token': 'apptoken',
    },
  })
    .post('/resource')
    .reply(200, { status: 'ok' })

  const d = domain.create()

  d.run(function() {
    mikealRequest({
      method: 'POST',
      uri: 'http://example.test/resource',
      headers: {
        'X-App-Token': 'apptoken',
      },
    })
  })

  d.once('error', function(err) {
    t.ok(err.message.match(/No match/))
    t.end()
  })
})

test('matches request header with regular expression', t => {
  nock('http://example.test', {
    reqheaders: {
      'X-My-Super-Power': /.+/,
    },
  })
    .post('/superpowers')
    .reply(200, { status: 'ok' })

  mikealRequest(
    {
      method: 'POST',
      uri: 'http://example.test/superpowers',
      headers: {
        'X-My-Super-Power': 'mullet growing',
      },
    },
    function(err, res, body) {
      t.strictEqual(err, null)
      t.equal(body, '{"status":"ok"}')
      t.end()
    }
  )
})

test('request header satisfies the header function', t => {
  nock('http://example.test', {
    reqheaders: {
      'X-My-Super-Power': function(value) {
        return value === 'mullet growing'
      },
    },
  })
    .post('/superpowers')
    .reply(200, { status: 'ok' })

  mikealRequest(
    {
      method: 'POST',
      uri: 'http://example.test/superpowers',
      headers: {
        'X-My-Super-Power': 'mullet growing',
      },
    },
    function(err, res, body) {
      t.strictEqual(err, null)
      t.equal(body, '{"status":"ok"}')
      t.end()
    }
  )
})

test("done fails when specified request header doesn't match regular expression", t => {
  nock('http://example.test', {
    reqheaders: {
      'X-My-Super-Power': /Mullet.+/,
    },
  })
    .post('/resource')
    .reply(200, { status: 'ok' })

  const d = domain.create()

  d.run(function() {
    mikealRequest({
      method: 'POST',
      uri: 'http://example.test/superpowers',
      headers: {
        'X-My-Super-Power': 'mullet growing',
      },
    })
  })

  d.once('error', function(err) {
    t.ok(err.message.match(/No match/))
    t.end()
  })
})

test("done fails when specified request header doesn't satisfy the header function", t => {
  nock('http://example.test', {
    reqheaders: {
      'X-My-Super-Power': function(value) {
        return value === 'Mullet Growing'
      },
    },
  })
    .post('/resource')
    .reply(200, { status: 'ok' })

  const d = domain.create()

  d.run(function() {
    mikealRequest({
      method: 'POST',
      uri: 'http://example.test/superpowers',
      headers: {
        'X-My-Super-Power': 'mullet growing',
      },
    })
  })

  d.once('error', function(err) {
    t.ok(err.message.match(/No match/))
    t.end()
  })
})

test('done does not fail when specified request header is not missing', t => {
  nock('http://example.test', {
    reqheaders: {
      'X-App-Token': 'apptoken',
      'X-Auth-Token': 'apptoken',
    },
  })
    .post('/resource')
    .reply(200, { status: 'ok' })

  mikealRequest(
    {
      method: 'POST',
      uri: 'http://example.test/resource',
      headers: {
        'X-App-Token': 'apptoken',
        'X-Auth-Token': 'apptoken',
      },
    },
    function(err, res, body) {
      t.type(err, 'null')
      t.equal(res.statusCode, 200)
      t.end()
    }
  )
})

test('done fails when specified bad request header is present', t => {
  nock('http://example.test', {
    badheaders: ['cookie'],
  })
    .post('/resource')
    .reply(200, { status: 'ok' })

  const d = domain.create()

  d.run(function() {
    mikealRequest({
      method: 'POST',
      uri: 'http://example.test/resource',
      headers: {
        Cookie: 'cookie',
      },
    })
  })

  d.once('error', function(err) {
    t.ok(err.message.match(/No match/))
    t.end()
  })
})

test('get correct filtering with scope and request headers filtering', t => {
  const responseText = 'OK!'
  const responseHeaders = { 'Content-Type': 'text/plain' }
  const requestHeaders = { host: 'a.subdomain.of.google.com' }

  const scope = nock('http://a.subdomain.of.google.com', {
    filteringScope: function(scope) {
      return /^http:\/\/.*\.google\.com/.test(scope)
    },
  })
    .get('/somepath')
    .reply(200, responseText, responseHeaders)

  let dataCalled = false
  const host = 'some.other.subdomain.of.google.com'
  const req = http.get(
    {
      host,
      method: 'GET',
      path: '/somepath',
      port: 80,
    },
    function(res) {
      res.on('data', function(data) {
        dataCalled = true
        t.equal(data.toString(), responseText)
      })
      res.on('end', function() {
        t.true(dataCalled)
        scope.done()
        t.end()
      })
    }
  )

  t.equivalent(req._headers, { host: requestHeaders.host })
})

test('mocking succeeds even when mocked and specified request header names have different cases', t => {
  nock('http://example.test', {
    reqheaders: {
      'x-app-token': 'apptoken',
      'x-auth-token': 'apptoken',
    },
  })
    .post('/resource')
    .reply(200, { status: 'ok' })

  mikealRequest(
    {
      method: 'POST',
      uri: 'http://example.test/resource',
      headers: {
        'X-App-TOKEN': 'apptoken',
        'X-Auth-TOKEN': 'apptoken',
      },
    },
    function(err, res, body) {
      t.type(err, 'null')
      t.equal(res.statusCode, 200)
      t.end()
    }
  )
})

// https://github.com/nock/nock/issues/966
test('mocking succeeds when mocked and specified request headers have falsy values', t => {
  nock('http://example.test', {
    reqheaders: {
      'x-foo': 0,
    },
  })
    .post('/resource')
    .reply(200, { status: 'ok' })

  mikealRequest(
    {
      method: 'POST',
      uri: 'http://example.test/resource',
      headers: {
        'X-Foo': 0,
      },
    },
    function(err, res, body) {
      t.error(err)
      t.equal(res.statusCode, 200)
      t.end()
    }
  )
})

test('mocking succeeds even when host request header is not specified', t => {
  nock('http://example.test')
    .post('/resource')
    .reply(200, { status: 'ok' })

  mikealRequest(
    {
      method: 'POST',
      uri: 'http://example.test/resource',
      headers: {
        'X-App-TOKEN': 'apptoken',
        'X-Auth-TOKEN': 'apptoken',
      },
    },
    function(err, res, body) {
      t.type(err, 'null')
      t.equal(res.statusCode, 200)
      t.end()
    }
  )
})

// https://github.com/nock/nock/issues/158
test('mikeal/request with strictSSL: true', t => {
  nock('https://example.test')
    .post('/what')
    .reply(200, { status: 'ok' })

  mikealRequest(
    {
      method: 'POST',
      uri: 'https://example.test/what',
      strictSSL: true,
    },
    function(err, res, body) {
      t.type(err, 'null')
      t.equal(res && res.statusCode, 200)
      t.end()
    }
  )
})

test('.setNoDelay', t => {
  nock('http://example.test')
    .get('/yay')
    .reply(200, 'Hi')

  const req = http.request(
    {
      host: 'example.test',
      path: '/yay',
      port: 80,
    },
    function(res) {
      t.equal(res.statusCode, 200)
      res.on('end', t.end.bind(t))
      // Streams start in 'paused' mode and must be started.
      // See https://nodejs.org/api/stream.html#stream_class_stream_readable
      res.resume()
    }
  )

  req.setNoDelay(true)

  req.end()
})

test('match basic authentication header', t => {
  const username = 'testuser'
  const password = 'testpassword'
  const authString = `${username}:${password}`

  const expectedAuthHeader = `Basic ${Buffer.from(authString).toString(
    'base64'
  )}`

  const scope = nock('http://example.test')
    .get('/')
    .matchHeader('Authorization', val => val === expectedAuthHeader)
    .reply(200, 'Hello World!')

  http.get(
    {
      host: 'example.test',
      path: '/',
      port: 80,
      auth: authString,
    },
    function(res) {
      res.setEncoding('utf8')
      t.equal(res.statusCode, 200)

      res.on('data', function(data) {
        t.equal(data, 'Hello World!')
      })

      res.on('end', function() {
        scope.done()
        t.end()
      })
    }
  )
})

test('request emits socket', t => {
  nock('http://example.test')
    .get('/')
    .reply(200, 'hey')

  const req = http.get('http://example.test')
  req.once('socket', function(socket) {
    t.equal(this, req)
    t.type(socket, Object)
    t.type(socket.getPeerCertificate(), 'string')
    t.end()
  })
})

test('socket emits connect and secureConnect', t => {
  t.plan(3)

  nock('http://example.test')
    .post('/')
    .reply(200, 'hey')

  const req = http.request({
    host: 'example.test',
    path: '/',
    method: 'POST',
  })

  req.on('socket', function(socket) {
    socket.once('connect', function() {
      req.end()
      t.ok(true)
    })
    socket.once('secureConnect', function() {
      t.ok(true)
    })
  })

  req.once('response', function(res) {
    res.setEncoding('utf8')
    res.on('data', function(d) {
      t.equal(d, 'hey')
    })
  })
})

test('socket setKeepAlive', t => {
  nock('http://example.test')
    .get('/')
    .reply(200, 'hey')

  const req = http.get('http://example.test')
  req.once('socket', function(socket) {
    socket.setKeepAlive(true)
    t.end()
  })
})

test('abort destroys socket', t => {
  nock('http://example.test')
    .get('/')
    .reply(200, 'hey')

  const req = http.get('http://example.test')
  req.once('error', function() {
    // ignore
  })
  req.once('socket', function(socket) {
    req.abort()
    t.ok(socket.destroyed)
    t.end()
  })
})

test('hyperquest works', t => {
  nock('http://example.test')
    .get('/somepath')
    .reply(200, 'Yay hyperquest!')

  const req = hyperquest('http://example.test/somepath')
  let reply = ''
  req.on('data', function(d) {
    reply += d
  })
  req.once('end', function() {
    t.equals(reply, 'Yay hyperquest!')
    t.end()
  })
})

test('match domain using regexp', t => {
  nock(/regexexample\.test/)
    .get('/resources')
    .reply(200, 'Match regex')

  mikealRequest.get('http://regexexample.test/resources', function(
    err,
    res,
    body
  ) {
    t.type(err, 'null')
    t.equal(res.statusCode, 200)
    t.equal(body, 'Match regex')

    t.end()
  })
})

test('match domain using regexp with path as callback (issue-1137)', t => {
  nock.cleanAll()
  nock(/.*/)
    .get(() => true)
    .reply(200, 'Match regex')

  mikealRequest.get('http://example.test/resources', function(err, res, body) {
    t.type(err, 'null')
    t.equal(res.statusCode, 200)
    t.equal(body, 'Match regex')
    t.end()
  })
})

test('match multiple interceptors with regexp domain (issue-508)', t => {
  nock.cleanAll()
  nock(/chainregex/)
    .get('/')
    .reply(200, 'Match regex')
    .get('/')
    .reply(500, 'Match second intercept')

  mikealRequest.get('http://chainregex.test', function(err, res, body) {
    t.type(err, 'null')
    t.equal(res.statusCode, 200)
    t.equal(body, 'Match regex')

    mikealRequest.get('http://chainregex.test', function(err, res, body) {
      t.type(err, 'null')
      t.equal(res.statusCode, 500)
      t.equal(body, 'Match second intercept')

      t.end()
    })
  })
})

test('match domain using intercept callback', t => {
  const validUrl = ['/cats', '/dogs']

  nock('http://example.test')
    .get(function(uri) {
      return validUrl.indexOf(uri) >= 0
    })
    .reply(200, 'Match intercept')
    .get('/cats')
    .reply(200, 'Match intercept 2')

  mikealRequest.get('http://example.test/cats', function(err, res, body) {
    t.type(err, 'null')
    t.equal(res.statusCode, 200)
    t.equal(body, 'Match intercept')

    // This one should match the second .get()
    mikealRequest.get('http://example.test/cats', function(err, res, body) {
      t.type(err, 'null')
      t.equal(res.statusCode, 200)
      t.equal(body, 'Match intercept 2')
      t.end()
    })
  })
})

test('match path using regexp', t => {
  nock('http://example.test')
    .get(/regex$/)
    .reply(200, 'Match regex')

  mikealRequest.get('http://example.test/resources/regex', function(
    err,
    res,
    body
  ) {
    t.type(err, 'null')
    t.equal(res.statusCode, 200)
    t.equal(body, 'Match regex')
    t.end()
  })
})

test('match path using function', t => {
  const path = '/match/uri/function'
  const options = {
    hostname: 'example.test',
    path,
  }
  const uriFunction = function(uri) {
    return uri === path
  }

  nock(`http://${options.hostname}`)
    .delete(uriFunction)
    .reply(200, 'Match DELETE')
    .get(uriFunction)
    .reply(200, 'Match GET')
    .head(uriFunction)
    .reply(200, 'Match HEAD')
    .merge(uriFunction)
    .reply(200, 'Match MERGE')
    .options(uriFunction)
    .reply(200, 'Match OPTIONS')
    .patch(uriFunction)
    .reply(200, 'Match PATCH')
    .post(uriFunction)
    .reply(200, 'Match POST')
    .put(uriFunction)
    .reply(200, 'Match PUT')

  options.method = 'POST'
  http
    .request(options, function(res) {
      res.setEncoding('utf8')
      t.equal(res.statusCode, 200)
      let body = ''
      res.on('data', function(data) {
        body += data
      })
      res.on('end', function() {
        t.equal(body, `Match ${options.method}`)

        options.method = 'GET'
        http
          .request(options, function(res) {
            res.setEncoding('utf8')
            t.equal(res.statusCode, 200)
            let body = ''
            res.on('data', function(data) {
              body += data
            })
            res.on('end', function() {
              t.equal(body, `Match ${options.method}`)

              options.method = 'OPTIONS'
              options.path = '/no/match'
              http
                .request(options)
                .on('error', e => {
                  t.similar(e.toString(), /Error: Nock: No match for request/)
                  t.end()
                })
                .end()
            })
          })
          .end()
      })
    })
    .end()
})

test('isDone() must consider repeated responses', t => {
  const scope = nock('http://example.test')
    .get('/')
    .times(2)
    .reply(204)

  function makeRequest(callback) {
    const req = http.request(
      {
        host: 'example.test',
        path: '/',
        port: 80,
      },
      function(res) {
        t.equal(res.statusCode, 204)
        res.on('end', callback)
        // Streams start in 'paused' mode and must be started.
        // See https://nodejs.org/api/stream.html#stream_class_stream_readable
        res.resume()
      }
    )
    req.end()
  }

  t.notOk(scope.isDone(), 'should not be done before all requests')
  makeRequest(function() {
    t.notOk(scope.isDone(), 'should not yet be done after the first request')
    makeRequest(function() {
      t.ok(scope.isDone(), 'should be done after the two requests are made')
      scope.done()
      t.end()
    })
  })
})

test('you must setup an interceptor for each request', t => {
  const scope = nock('http://example.test')
    .get('/hey')
    .reply(200, 'First match')

  mikealRequest.get('http://example.test/hey', function(error, res, body) {
    t.equal(res.statusCode, 200)
    t.equal(body, 'First match', 'should match first request response body')

    mikealRequest.get('http://example.test/hey', function(error, res, body) {
      t.equal(
        error && error.toString(),
        `Error: Nock: No match for request ${JSON.stringify(
          {
            method: 'GET',
            url: 'http://example.test/hey',
            headers: { host: 'example.test' },
          },
          null,
          2
        )}`
      )
      scope.done()
      t.end()
    })
  })
})

test('replyWithError returns an error on request', t => {
  const scope = nock('http://example.test')
    .post('/echo')
    .replyWithError('Service not found')

  const req = http.request({
    host: 'example.test',
    method: 'POST',
    path: '/echo',
    port: 80,
  })

  // An error should have have been raised
  req.on('error', function(e) {
    scope.done()
    t.equal(e.message, 'Service not found')
    t.end()
  })

  req.end()
})

test('replyWithError allows json response', t => {
  const scope = nock('http://example.test')
    .post('/echo')
    .replyWithError({ message: 'Service not found', code: 'test' })

  const req = http.request({
    host: 'example.test',
    method: 'POST',
    path: '/echo',
    port: 80,
  })

  // An error should have have been raised
  req.on('error', function(e) {
    scope.done()
    t.equal(e.message, 'Service not found')
    t.equal(e.code, 'test')
    t.end()
  })

  req.end()
})

test('no content type provided', t => {
  const scope = nock('http://example.test')
    .replyContentLength()
    .post('/httppost', function() {
      return true
    })
    .reply(401, '')

  http
    .request(
      {
        host: 'example.test',
        path: '/httppost',
        method: 'POST',
        headers: {},
      },
      function(res) {
        res.on('data', function() {})
        res.once('end', function() {
          scope.done()
          t.ok(true)
          t.end()
        })
      }
    )
    .end('WHAA')
})

// https://github.com/nock/nock/issues/835
test('match domain and path using regexp', t => {
  nock.cleanAll()
  const imgResponse = 'Matched Images Page'

  const scope = nock(/example/)
    .get(/img/)
    .reply(200, imgResponse)

  mikealRequest.get('http://example.test/imghp?hl=en', function(
    err,
    res,
    body
  ) {
    scope.done()
    t.type(err, 'null')
    t.equal(res.statusCode, 200)
    t.equal(body, imgResponse)
    t.end()
  })
})

test('multiple interceptors override headers from unrelated request', t => {
  nock.cleanAll()

  nock.define([
    {
      scope: 'https://example.test:443',
      method: 'get',
      path: '/bar',
      reqheaders: {
        'x-foo': 'bar',
      },
      status: 200,
      response: {},
    },
    {
      scope: 'https://example.test:443',
      method: 'get',
      path: '/baz',
      reqheaders: {
        'x-foo': 'baz',
      },
      status: 200,
      response: {},
    },
  ])

  mikealRequest(
    {
      url: 'https://example.test/bar',
      headers: {
        'x-foo': 'bar',
      },
    },
    function(err, res, body) {
      t.error(err)
      t.equal(res.statusCode, 200)

      mikealRequest.get(
        {
          url: 'https://example.test/baz',
          headers: {
            'x-foo': 'baz',
          },
        },
        function(err, res, body) {
          t.error(err)
          t.equal(res.statusCode, 200)
          t.end()
        }
      )
    }
  )
})

// https://github.com/nock/nock/issues/1003
test('correctly parse request without specified path', t => {
  nock.cleanAll()

  const scope1 = nock('https://example.test')
    .get('')
    .reply(200)

  https
    .request({ hostname: 'example.test' }, function(res) {
      t.equal(res.statusCode, 200)
      res.on('data', function() {})
      res.on('end', function() {
        scope1.done()
        t.end()
      })
    })
    .end()
})

test('data is sent with flushHeaders', t => {
  nock.cleanAll()

  const scope1 = nock('https://example.test')
    .get('')
    .reply(200, 'this is data')

  https
    .request({ hostname: 'example.test' }, function(res) {
      t.equal(res.statusCode, 200)
      res.on('data', function(data) {
        t.equal(data.toString(), 'this is data')
      })
      res.on('end', function() {
        scope1.done()
        t.end()
      })
    })
    .flushHeaders()
})

test('should log matching', async t => {
  const messages = []

  const scope = nock('http://example.test')
    .get('/')
    .reply(200, 'Hello, World!')
    .log(message => messages.push(message))

  await got('http://example.test/')

  t.deepEqual(messages, [
    'matching http://example.test:80/ to GET http://example.test:80/: true',
  ])

  scope.done()
})

test('teardown', t => {
  let leaks = Object.keys(global).splice(globalCount, Number.MAX_VALUE)

  leaks = leaks.filter(function(key) {
    return acceptableLeaks.indexOf(key) == -1
  })

  t.deepEqual(leaks, [], 'No leaks')
  t.end()
})
